#!/usr/bin/env python

import asyncio
import asyncio.trsock
import socket
import time
from picows import ws_create_server, WSFrame, WSTransport, WSListener, WSMsgType, WSUpgradeRequest, WSCloseCode
import threading
import logging
from queue import Queue
import argparse

from pytunneler.utils import commands, network, packet, host_hex

parser = argparse.ArgumentParser(description='pytunneler Websocket server')

parser.add_argument('address', type=str, nargs='?',
                    help='websocket server address', default='0.0.0.0:8321')

parser.add_argument('--password', type=str,
                    help='password')

class WebsocketServer(WSListener):
    def __init__(self, ip, port, password):
        self.tcp_thread = None
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.tcp_servers: dict[str, TCPServer] = dict()
        self.local_remote: dict[str, str] = dict()
        self.ip = ip
        self.port = port
        self.password = password


    def start_tcp_server(self, transport, local_ip, local_port, target_ip, target_port):
        thread = TCPServer(transport, self, local_ip, int(local_port), target_ip, int(target_port))
        self.tcp_servers[f"{target_ip}:{target_port}"] = thread
        self.local_remote[f"{local_ip}:{local_port}"] = f"{target_ip}:{target_port}"
        thread.start()

    def send_callback(self, transport, message):
        command_callback_packet = packet.CommandCallbackPacket(message)
        network.websocket_send(transport, command_callback_packet.create_packet())

    def handle_packet(self, raw_packet, transport: WSTransport):
        packet_ = packet.Packet.bytes2packet(raw_packet)
        if isinstance(packet_, packet.CommandPacket):
            command = packet_.command
            for command_type in commands.CommandTypes:
                if command.split()[0] == command_type.trigger:
                    context = commands.CommandContext(command.split()[1:], transport, self)
                    message = command_type.on_command(context)
                    self.send_callback(transport, message)
                    break
            else:
                print(f"Unknown command: {command.split()[0]}")
                self.send_callback(transport, f"Unknown command: {command.split()[0]}")
        elif isinstance(packet_, packet.BinaryPacket):
            try:
                ip, port = host_hex.hex2host(packet_.host).split(":")
                target_ip, target_port = host_hex.hex2host(packet_.target_host).split(":")
                remote_address = self.local_remote[f"{ip}:{port}"]

                if remote_address not in self.tcp_servers:
                    logging.error(f"No TCP server found for {remote_address}")
                    return

                tcp_server = self.tcp_servers[remote_address]
                client_key = f"{target_ip}:{target_port}"

                if client_key not in tcp_server.clients:
                    logging.error(f"No client connection found for {client_key}")
                    return

                client_handler = tcp_server.clients[client_key]
                network.tcp_send(client_handler.client_socket, packet_.data)
            except Exception as e:
                logging.error(f"Error handling binary packet: {e}")


    def check_password(self, transport, frame: WSFrame):
        try:
            password_packet = frame.get_payload_as_bytes()
            password_packet = packet.Packet.bytes2packet(password_packet)
            if isinstance(password_packet, packet.PasswordPacket):
                if self.password != password_packet.password:
                    self.send_callback(transport, "Wrong password")
                    time.sleep(1) # 给客户端时间反应，不然直接退出了信息看不到
                    return False
            else:
                self.send_callback(transport, "Wrong packet")
                time.sleep(1)
                return False
        except TimeoutError:
            self.send_callback(transport, "Timeout for password")
            time.sleep(1)
            return False
        return True




    class ServerListener(WSListener):
        def __init__(self, server):
            self.server = server
            self.password_checked = False
            super().__init__()

        def on_ws_connected(self, transport: WSTransport):
            atransport: asyncio.Transport = transport.underlying_transport
            transport_socket: asyncio.trsock.TransportSocket = atransport.get_extra_info('socket')
            peer = transport_socket.getpeername()
            logging.info(f"Connected by {peer[0]}:{peer[1]}")


        def on_ws_frame(self, transport: WSTransport, frame: WSFrame):
            if frame.msg_type == WSMsgType.PING:
                transport.send_pong(frame.get_payload_as_bytes())
            elif frame.msg_type == WSMsgType.CLOSE:
                transport.send_close(frame.get_close_code(), frame.get_close_message())
                transport.disconnect()
            else:
                if not self.password_checked:
                    if server.check_password(transport, frame):
                        self.password_checked = True
                        return
                    else:
                        transport.send_close(WSCloseCode.INVALID_TEXT)
                        transport.disconnect()
                        return
                raw_packet = frame.get_payload_as_bytes()
                if not raw_packet:
                    transport.send_close(WSCloseCode.BAD_GATEWAY)
                    transport.disconnect()
                self.server.handle_packet(raw_packet, transport)


    def listener_factory(self, r: WSUpgradeRequest):
        return self.ServerListener(self)

    async def main(self):
        server: asyncio.Server = await ws_create_server(self.listener_factory, self.ip, self.port)
        for s in server.sockets:
            logging.info(f"Websocket server is running on {s.getsockname()}")
        await server.serve_forever()



    def shutdown(self):
        """Shutdown all TCP servers and cleanup resources"""
        for tcp_server in self.tcp_servers.values():
            tcp_server.shutdown.set()
        self.loop.stop()

    def run(self):
        try:
            return self.loop.run_until_complete(self.main())
        except KeyboardInterrupt:
            self.shutdown()
        except Exception as e:
            logging.error(f"Server error: {e}")
            self.shutdown()

class ConnectionHandler(threading.Thread):
    def __init__(self, client_socket, client_address, websocket, local_hex):
        self.client_socket = client_socket
        self.client_address = client_address
        self.websocket = websocket
        self.local_hex = local_hex
        self.client_hex = host_hex.host2hex(f"{client_address[0]}:{client_address[1]}")
        super().__init__()

    def run(self):
        with self.client_socket:
            logging.info(f"Connected by {self.client_address}")
            connected_packet = packet.ConnectedPacket(self.client_hex, self.local_hex)
            network.websocket_send(self.websocket, connected_packet.create_packet())
            time.sleep(0.5)

            while True:
                data = network.tcp_recv(self.client_socket)
                if data is None:
                    logging.info(f"Connection closed by {self.client_address}")
                    break
                binary_packet = packet.BinaryPacket(self.client_hex, self.local_hex, data)
                network.websocket_send(self.websocket, binary_packet.create_packet())

class TCPServer(threading.Thread):
    def __init__(self, websocket: WSTransport, server: WebsocketServer, local_ip, local_port, host, port):
        self.packet_queue = Queue()
        self.local_ip = local_ip
        self.local_port = local_port
        self.host = host
        self.port = port
        self.server = server
        self.server_socket = None
        self.websocket = websocket
        self.clients = {}
        self.shutdown = threading.Event()
        self.local_hex = host_hex.host2hex(f"{local_ip}:{local_port}")
        super().__init__()

    def run(self):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as server_socket:
            server_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind((self.host, self.port))
            server_socket.listen(100)  # Increased backlog
            self.server_socket = server_socket
            logging.info(f"TCP server is running on {self.host}:{self.port}")

            while not self.shutdown.is_set():
                try:
                    client_socket, client_address = server_socket.accept()
                    client_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

                    # Create and start new connection handler thread
                    handler = ConnectionHandler(client_socket, client_address, self.websocket, self.local_hex)
                    handler.daemon = True  # Don't block program exit
                    handler.start()

                    # Track client connection
                    client_host, client_port = client_address
                    self.clients[f"{client_host}:{client_port}"] = handler

                except OSError as e:
                    if not self.shutdown.is_set():
                        logging.error(f"TCP server error: {e}")
                    break

            # Cleanup all client connections
            for handler in list(self.clients.values()):
                try:
                    handler.client_socket.close()
                except:
                    pass
                handler.join(timeout=1)

            # Close server socket
            try:
                self.server_socket.close()
            except:
                pass

if __name__ == "__main__":
    args = parser.parse_args()
    ip, port = args.address.split(":")
    password = ""
    if args.password:
        password = args.password
    server = WebsocketServer(ip, port, password)
    server.run()



import asyncio
import logging
import socket
from picows import WSTransport, WSMsgType
BUFFER = 65536  # Increased from 8KB to 64KB


def tcp_recv(client_socket):
    try:
        data = client_socket.recv(BUFFER)
        if data:
            logging.debug(f"Received {len(data)} bytes")
        if not data:
            return None
        logging.debug("TCP server received: " + str(data))
        return data
    except Exception as e:
        logging.error(f"Error receiving TCP data: {e}")
        return None

def tcp_send(client_socket, data):
    try:
        client_socket.sendall(data)
    except Exception as e:
        logging.error(f"Error sending TCP data: {e}")
        return False
    return True

def websocket_send(transport: WSTransport, data):
    try:
        transport.send(WSMsgType.BINARY, data)
    except Exception as e:
        logging.debug(f"Error sending WebSocket data: {e}")


def try_port(port):
    # Source: https://stackoverflow.com/a/43271125/24191134
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        result = False
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind(("127.0.0.1", port))
            result = True
        except socket.error as e:
            pass
        finally:
            sock.close()
        return result

def scan_port():
    available_ports = []
    for i in range(1, 65535):
        if try_port(i):
            available_ports.append(i)
    return available_ports



from __future__ import annotations

from websockets import WebSocketServerProtocol

from pytunneler.server.server import WebsocketServer
from tabulate import tabulate
import traceback
from pytunneler.utils import network


class CommandContext:
    def __init__(self, args, websocket: WebSocketServerProtocol, server: WebsocketServer):
        self.args = args
        self.server = server
        self.websocket = websocket

class Command:
    def __init__(self):
        self.name = "Reversed Command"
        self.description = "This is a Reversed Command"
        self.trigger = "reversed"

    @staticmethod
    def on_command(context: CommandContext):
        return "Hello World"

class LsCommandCommand(Command):
    def __init__(self):
        super().__init__()
        self.name = "List of Commands"
        self.description = "List of Commands"
        self.trigger = "lsc"

    @staticmethod
    def on_command(context: CommandContext):
        result = []
        for command in CommandTypes:
            result.append(
                [command.name, command.description, command.trigger]
            )
        return tabulate(result, headers=["name", "description", "trigger"])

class LsPortCommand(Command):
    def __init__(self):
        super().__init__()
        self.name = "List of Ports"
        self.description = "List of Ports"
        self.trigger = "lsp"

    @staticmethod
    def on_command(context: CommandContext):
        ports = network.scan_port()

        result = []
        start = ports[0]
        end = ports[0]

        for i in range(1, len(ports)): # Compress ports to ranges and numbers
            if ports[i] == end + 1:
                end = ports[i]
            else:
                if start == end:
                    result.append(f"{start}")
                else:
                    result.append(f"{start}-{end}")
                start = ports[i]
                end = ports[i]

        if start == end:
            result.append(f"{start}")
        else:
            result.append(f"{start}-{end}")

        return f'Available Ports: {" ".join(result)}'

class TcpTunnelCommand(Command):
    def __init__(self):
        super().__init__()
        self.name = "TcpTunneling"
        self.description = "make a tcp socket tunnel to server"
        self.trigger = "tcptunnel"

    @staticmethod
    def usage():
        return "Usage: tcptunnel [localhost]:<port> :[target_port]"

    @staticmethod
    def on_command(context: CommandContext):
        if len(context.args) != 2:
            return TcpTunnelCommand.usage()
        try:
            local_ip, local_port = context.args[0].split(":")
            target_ip, target_port = context.args[1].split(":")
            if not local_ip: local_ip = "127.0.0.1"
            if not target_ip: target_ip = "0.0.0.0"
            context.server.start_tcp_server(context.websocket, local_ip, local_port, target_ip, target_port)
            return f"tunneling {local_ip}:{local_port} to {target_ip}:{target_port}"
        except Exception:
            return traceback.format_exc()

CommandTypes : list[type[Command]] = [
    LsCommandCommand(), LsPortCommand(), TcpTunnelCommand()
]




from typing import List, Type


#  websocket层传输的包定义
DATA_LENGTH = 15
#  前15字节保留为包信息



class Packet:
    def __init__(self, data: bytes):
        self.data = data
        self.addition_data = b""

    @staticmethod
    def get_type_magic_number() -> bytes:
        return b"\x01"

    def create_packet(self):
        if len(self.addition_data) > 14:
            raise RuntimeError(f"data has more than 9 bytes: {self.addition_data}")
        length_byte = len(self.addition_data).to_bytes(1, 'big')
        tmp = self.addition_data + (14 - len(self.addition_data)) * b"\x00"
        return self.get_type_magic_number() + length_byte + tmp + self.data

    @classmethod
    def from_packet(cls, bytes):
        if bytes[0:1] == cls.get_type_magic_number():
            data = bytes[16:]
            return cls(data)
        else:
            raise RuntimeError("packet type don't match")

    @staticmethod
    def bytes2packet(bytes: bytes):
        for type_ in PacketTypes:
            if bytes[0:1] == type_.get_type_magic_number():
                return type_.from_packet(bytes)
        raise RuntimeError("no such packet type")

class ConnectedPacket(Packet):
    def __init__(self, host: bytes, target_host: bytes):
        super().__init__(b'')
        self.host = host
        self.target_host = target_host
        self.addition_data = host + target_host

    @staticmethod
    def get_type_magic_number() -> bytes:
        return b"\x02"

    @classmethod
    def from_packet(cls, bytes):
        if bytes[0:1] == cls.get_type_magic_number():
            length = bytes[1]
            host = bytes[2:2 + int(length / 2)]
            target_host = bytes[2 + int(length / 2): length + 2]
            return cls(host, target_host)
        else:
            raise RuntimeError("packet type don't match")

class ShutdownPacket(Packet):
    def __init__(self, host):
        super().__init__(b'')
        self.addition_data = host

    @staticmethod
    def get_type_magic_number() -> bytes:
        return b"\x03"

    @classmethod
    def from_packet(cls, bytes):
        if bytes[0:1] == cls.get_type_magic_number():
            length = bytes[1]
            host = bytes[2:2 + length]
            return cls(host)
        else:
            raise RuntimeError("packet type don't match")

class BinaryPacket(Packet):
    def __init__(self, host: bytes, target_host: bytes, data: bytes):
        super().__init__(data)
        self.host = host
        self.target_host = target_host
        self.addition_data = host + target_host

    @staticmethod
    def get_type_magic_number() -> bytes:
        return b"\x04"

    @classmethod
    def from_packet(cls, bytes):
        if bytes[0:1] == cls.get_type_magic_number():
            length = bytes[1]
            host = bytes[2:2 + int(length / 2)]
            target_host = bytes[2 + int(length / 2): length + 2]
            data = bytes[16:]
            return cls(host, target_host, data)
        else:
            raise RuntimeError("packet type don't match")

class CommandPacket(Packet):
    def __init__(self, command: str):
        self.command = command
        super().__init__(command.encode())

    @staticmethod
    def get_type_magic_number() -> bytes:
        return b"\x05"

    @classmethod
    def from_packet(cls, bytes):
        if bytes[0:1] == cls.get_type_magic_number():
            command = bytes[16:]
            return cls(command.decode())
        else:
            raise RuntimeError("packet type don't match")

class CommandCallbackPacket(Packet):
    def __init__(self, message: str):
        self.message = message
        super().__init__(message.encode())

    @staticmethod
    def get_type_magic_number() -> bytes:
        return b"\x06"

    @classmethod
    def from_packet(cls, bytes):
        if bytes[0:1] == cls.get_type_magic_number():
            message = bytes[16:]
            return cls(message.decode())
        else:
            raise RuntimeError("packet type don't match")

class PasswordPacket(Packet):
    def __init__(self, password: str):
        self.password = password
        super().__init__(password.encode())

    @staticmethod
    def get_type_magic_number() -> bytes:
        return b"\x07"

    @classmethod
    def from_packet(cls, bytes):
        if bytes[0:1] == cls.get_type_magic_number():
            password = bytes[16:]
            return cls(password.decode())
        else:
            raise RuntimeError("packet type don't match")


PacketTypes : List[Type[Packet]] = [
    Packet, ConnectedPacket, ShutdownPacket, BinaryPacket, CommandPacket, CommandCallbackPacket, PasswordPacket
]



#!/usr/bin/env python

import asyncio
import logging
from queue import Queue
from picows import ws_connect, WSFrame, WSTransport, WSListener, WSMsgType, WSCloseCode, WSUpgradeRequest
import threading
import aioconsole
import socket
import argparse

from pytunneler.utils import commands, network, packet, host_hex

logging.getLogger().setLevel(logging.INFO)

parser = argparse.ArgumentParser(description='pytunneler Websocket client')

parser.add_argument('address', type=str, nargs='?',
                    help='websocket server address', default='127.0.0.1:8321')

parser.add_argument('--password', type=str,
                    help='password')

parser.add_argument('--init_commands', type=str,
                    help='password')

class WebsocketClient:
    def __init__(self, ip, port, password, init_commands) -> None:
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.tcp_clients: dict[str, TCPClient] = dict()
        self.local_remote: dict[str, str] = dict()
        self.init_commands = init_commands
        self.ip = ip
        self.port = port
        self.password = password

    def send_command(self, transport: WSTransport, command):
        print(command)
        if not command.strip():
            print("> ", end='', flush=True)
            return
        for command_type in commands.CommandTypes:
            if command.split()[0] == command_type.trigger:  # 找到命令
                command_packet = packet.CommandPacket(command)
                network.websocket_send(transport, command_packet.create_packet())
                break
        else:
            print(f"Unknown command: {command.split()[0]}")


    async def handle_command_input(self, transport: WSTransport):
        """Handle user command and send it to the websocket"""
        print("> ", end='', flush=True)
        while True:
            command = await aioconsole.ainput("")
            self.send_command(transport, command)

    def start_tcp_client(self, transport, target_ip, target_port, ip, port):
        thread = TCPClient(transport, self, target_ip, int(target_port), ip, int(port))
        self.tcp_clients[f"{ip}:{port}"] = thread
        thread.start()


    def handle_packet(self, raw_packet, transport):
        packet_ = packet.Packet.bytes2packet(raw_packet)
        if isinstance(packet_, packet.CommandCallbackPacket):
            print(packet_.message)
            print("> ", end='', flush=True)
        elif isinstance(packet_, packet.ConnectedPacket):
            ip, port = host_hex.hex2host(packet_.host).split(":")
            target_ip, target_port = host_hex.hex2host(packet_.target_host).split(":")
            self.start_tcp_client(transport, target_ip, target_port, ip, port)

        elif isinstance(packet_, packet.BinaryPacket):
            ip, port = host_hex.hex2host(packet_.host).split(":")
            target_ip, target_port = host_hex.hex2host(packet_.target_host).split(":")
            client = self.tcp_clients[f"{ip}:{port}"]
            if not client.sock:
                print("error")
            network.tcp_send(client.sock, packet_.data)



    def send_password(self, transport):
        password_packet = packet.PasswordPacket(self.password)         # send password first
        network.websocket_send(transport, password_packet.create_packet())

    def send_init_commands(self, transport):
        for command in init_commands:
            self.send_command(transport, command)

    class ClientListener(WSListener):
        def on_ws_connected(self, transport: WSTransport):
            self.transport = transport
            client.send_password(transport)
            client.send_init_commands(transport)
            self.input_thread = threading.Thread(target=asyncio.run_coroutine_threadsafe, args=(client.handle_command_input(transport), client.loop))
            self.input_thread.start()


        def on_ws_frame(self, transport: WSTransport, frame: WSFrame):
            raw_packet = frame.get_payload_as_bytes()
            if not raw_packet:
                transport.send_close(WSCloseCode.BAD_GATEWAY)
                transport.disconnect()
                return
            client.handle_packet(raw_packet, transport)


        def on_ws_disconnected(self, transport):
            self.input_thread.join()



    async def main(self):
        logging.info(f"Websocket client connected to ws://{self.ip}:{self.port}")
        (_, client) = await ws_connect(self.ClientListener, f"ws://{self.ip}:{self.port}")
        client.client = self # todo: hacky way
        await client.transport.wait_disconnected()

    def run(self):
        return self.loop.run_until_complete(self.main())


class TCPClient(threading.Thread):
    def __init__(self, websocket: WSTransport, client: WebsocketClient, target_ip, target_port, ip, port):
        self.packet_queue = Queue()
        self.target_ip = target_ip
        self.target_port = target_port
        self.target_hex =host_hex.host2hex(f"{target_ip}:{target_port}")
        self.hex = host_hex.host2hex(f"{ip}:{port}")
        self.client = client
        self.websocket = websocket
        self.sock = None
        super().__init__()

    def run(self):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            self.sock = sock
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            sock.settimeout(5)
            sock.connect((self.target_ip, self.target_port))
            logging.info("Tcp client connected to " + str((self.target_ip, self.target_port)))
            while True:
                data = network.tcp_recv(sock)
                if not data:
                    logging.info("no data, shutdowning")
                    break
                binary_packet = packet.BinaryPacket(self.target_hex, self.hex, data)
                network.websocket_send(self.websocket, binary_packet.create_packet())
        exit(0)


if __name__ == "__main__":
    args = parser.parse_args()
    init_commands = []
    if args.init_commands:
        init_commands = args.init_commands.replace('[', "").replace(']', "").replace("\"", "").split(",")
    password = ""
    if args.password:
        password = args.password
    ip, port = args.address.split(":")
    client = WebsocketClient(ip, int(port), password, init_commands)
    client.run()



import socket

def host2hex(host):
    # Convert IP address to hash
    ip, port = host.split(":")
    ip = socket.inet_aton(ip)
    port = int(port).to_bytes(2, 'big')
    return ip + port

def hex2host(hash_):
    # Convert the IP address bytes back to a string
    ip = socket.inet_ntoa(hash_[:4])
    # Convert the port number bytes back to an integer
    port = int.from_bytes(hash_[4:], 'big')
    return f"{ip}:{port}"

